/*
 * CPU speed daemon - client library
 *
 * Copyright 2003 Jeremy Fitzhardinge <jeremy@goop.org>
 *
 * This software is licensed under the GNU General Public License.
 * See COPYING for details.
 */

#include <stddef.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/poll.h>

#include "speedfreq.h"
#include "speedfreq-int.h"

extern char version[];

static int sockfd = -1;

static int sendcred(int fd, void *data, size_t sz, int flags)
{
	struct ucred *uc;
	struct cmsghdr *cmsg;
	struct msghdr msg = { 0, };
	char buf[CMSG_SPACE(sizeof(struct ucred))];
	struct iovec iov;

	/* Send our credentials with each message.
	   Gak, what a mess... */
	msg.msg_control = buf;
	msg.msg_controllen = sizeof(buf);
	cmsg = CMSG_FIRSTHDR(&msg);
	cmsg->cmsg_level = SOL_SOCKET;
	cmsg->cmsg_type = SCM_CREDENTIALS;
	cmsg->cmsg_len = CMSG_LEN(sizeof(struct ucred));
	uc = (struct ucred *)CMSG_DATA(cmsg);
	msg.msg_controllen = cmsg->cmsg_len;

	uc->pid = getpid();
	uc->uid = getuid();
	uc->gid = getgid();

	iov.iov_base = data;
	iov.iov_len = sz;

	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;

	return sendmsg(fd, &msg, flags);
}

/* Send command to server; returns 0 on success, 1 on error */
static int sendcommand(int fd, struct sfi_request *cmd, const char *desc)
{
	int ret;

	/* send command */
	if (sendcred(fd, cmd, sizeof(*cmd), MSG_NOSIGNAL) != sizeof(*cmd)) {
		perror("sendmsg");
		ret = 1;
		goto out;
	}

	/* get result */
	if (recv(fd, cmd, sizeof(*cmd), MSG_NOSIGNAL) != sizeof(*cmd)) {
		perror("recv");
		ret = 1;
		goto out;
	}

	switch(cmd->cmd) {
	case OK:
		ret = 0;
		break;
	case ERROR:
		fprintf(stderr, "%s failed: %s\n", desc, cmd->string);
		ret = 1;
		break;
	default:
		fprintf(stderr, "%s: unexpected reply %d\n", desc, cmd->cmd);
		abort();
	}

  out:
	return ret;

}

static const struct policies {
	const char *name;
} policies[] = {
	[SF_POWERSAVE]	= { "powersave" },
	[SF_PERFORMANCE]= { "performance" },
	[SF_DYNAMIC]	= { "dynamic" },
	[SF_FIXED]	= { "fixed" },
};

const char *sf_policyname(enum sf_policy p) {
	if (p < 0 || p > sizeof(policies)/sizeof(*policies))
		return NULL;

	return policies[p].name;
}

static int do_connect(void)
{
	static const struct sockaddr_un addr = { AF_UNIX, SOCKPATH };
	static const int addrlen = offsetof(struct sockaddr_un, sun_path) + sizeof(SOCKPATH) - 1;
	int fd;
	const int passcred = 1;

	fd = socket(PF_UNIX, SOCK_DGRAM, 0);

	if (fd == -1) {
		return -1;
	}

	if (setsockopt(fd, SOL_SOCKET, SO_PASSCRED, &passcred, sizeof(passcred)) == -1)
		perror("setsockopt/PASSCRED");

	if (connect(fd, (const struct sockaddr *)&addr, addrlen) == -1) {
		close(fd);
		return -1;
	}

	return fd;
}

/* Connect to server - return 1 on error, 0 on success*/
int sf_connect(void)
{
	struct sfi_request ping = { .cmd = GETVERSION };
	int ret;

	if (sockfd == -1) {
		/* not already connected */
		sockfd = do_connect();

		if (sockfd == -1)
			return 1;
	}

	ret = sendcommand(sockfd, &ping, "Pinging server");

	if (ret == 1 ||
	    (ret == 0 && strncmp(ping.string, version, sizeof(ping.string)) != 0)) {
		close(sockfd);
		sockfd = -1;
		ret = 1;
	}

	return ret;
}

void sf_disconnect(void)
{
	if (sockfd != -1) {
		close(sockfd);
		sockfd = -1;
	}
}

char *sf_getversion(void)
{
	struct sfi_request cmd;
	char *ret = NULL;

	memset(&cmd, 0, sizeof(cmd));
	cmd.cmd = GETVERSION;

	if (!sendcommand(sockfd, &cmd, "Getting version"))
		ret = strdup(cmd.string);

	return ret;
}

int sf_getpolicy(struct sf_policyinfo *info)
{
	struct sfi_request cmd;
	int ret = 1;

	memset(&cmd, 0, sizeof(cmd));
	cmd.cmd = GETPOLICY;

	if (!sendcommand(sockfd, &cmd, "Getting policy info")) {
		*info = cmd.policy;
		ret = 0;
	}

	return ret;
}

int sf_pol_powersave(void)
{
	struct sfi_request cmd;

	memset(&cmd, 0, sizeof(cmd));
	cmd.cmd = SETPOLICY;
	cmd.policy.policy = SF_POWERSAVE;

	return sendcommand(sockfd, &cmd, "Setting powersave policy");
}

int sf_pol_performance(void)
{
	struct sfi_request cmd;

	memset(&cmd, 0, sizeof(cmd));
	cmd.cmd = SETPOLICY;
	cmd.policy.policy = SF_PERFORMANCE;

	return sendcommand(sockfd, &cmd, "Setting performance policy");
}

int sf_pol_dynamic(unsigned min, unsigned max)
{
	struct sfi_request cmd;

	memset(&cmd, 0, sizeof(cmd));
	cmd.cmd = SETPOLICY;
	cmd.policy.policy = SF_DYNAMIC;
	cmd.policy.dynamic.min = min;
	cmd.policy.dynamic.max = max;

	return sendcommand(sockfd, &cmd, "Setting dynamic policy");
}

int sf_pol_fixed(unsigned speed)
{
	struct sfi_request cmd;

	memset(&cmd, 0, sizeof(cmd));
	cmd.cmd = SETPOLICY;
	cmd.policy.policy = SF_FIXED;
	cmd.policy.fixed.current = speed;

	return sendcommand(sockfd, &cmd, "Setting fixed policy");
}

int sf_getcpuinfo(struct sf_cpuinfo *info)
{
	struct sfi_request cmd;
	int ret = 1;

	memset(&cmd, 0, sizeof(cmd));
	cmd.cmd = GETCPUINFO;

	if (!sendcommand(sockfd, &cmd, "Getting CPU info")) {
		ret = 0;
		*info = cmd.cpuinfo;
	}

	return ret;
}

int sf_shutdown(void)
{
	struct sfi_request cmd;

	memset(&cmd, 0, sizeof(cmd));
	cmd.cmd = SHUTDOWN;

	return sendcommand(sockfd, &cmd, "Shutting down server");
}


static int _mon_fd = -1;

int sf_get_mon_fd(void)
{
	return _mon_fd;
}

static inline int sf_set_mon_fd(int fd)
{
	return _mon_fd = fd;
}


static int prepare(struct sfi_request *cmd, struct pollfd *fds)
{
	int mon_fd = sf_get_mon_fd();

	if (mon_fd == -1) {
		mon_fd = sf_set_mon_fd(do_connect());
		if (mon_fd == -1)
			return -1;

		cmd->cmd = MONITOR;
		cmd->enable = 1;

		if (sendcommand(mon_fd, cmd, "Opening monitor connection")) {
			close(mon_fd);
			mon_fd = sf_set_mon_fd(-1);
			return -1;
		}

		shutdown(mon_fd, SHUT_WR); /* become RO */
		fds->fd = mon_fd;
		fds->events = POLLIN|POLLERR|POLLHUP;
	}

	assert(mon_fd != -1);
	assert(fds->fd == mon_fd);

	return mon_fd;
}

static void go_away(struct sfi_request *cmd)
{
	int mon_fd = sf_get_mon_fd();

	if (mon_fd != -1) {
		cmd->cmd = MONITOR;
		cmd->enable = 0;
		sendcommand(mon_fd, cmd, "Closing monitor connection");
		close(mon_fd);
		sf_set_mon_fd(-1);
	}
}

int sf_monitor(const struct sf_monitorcb *cb, int blocking)
{
	static struct pollfd fds;
	struct sfi_request cmd;
	int timeout;
	int mon_fd;

	memset(&cmd, 0, sizeof(cmd));
	if (cb == NULL) {
		go_away(&cmd);
		return -1;
	}

	if ((mon_fd = prepare(&cmd, &fds)) == -1)
		return -1;

	timeout = blocking ? -1 : 0;

	while(poll(&fds, 1, timeout) > 0) {
		int r;

		if (fds.revents & (POLLERR|POLLHUP))
			break;

		r = recv(mon_fd, &cmd, sizeof(cmd), MSG_NOSIGNAL);

		timeout = 0;

		if (r != sizeof(cmd))
			break;

		switch(cmd.cmd) {
		case GETCPUINFO:
			if (cb->cpuinfo)
				(*cb->cpuinfo)(&cmd.cpuinfo);
			break;

		case GETPOLICY:
			if (cb->policy)
				(*cb->policy)(&cmd.policy);
			break;

		case SHUTDOWN:
			/* server gone away */
			close(mon_fd);
			sf_set_mon_fd(-1);
			goto out;

		default:
			fprintf(stderr, "Unexpected monitor event: %d\n", cmd.cmd);
			break;
		}
	}
  out:
	return mon_fd;
}


int sf_poll(struct sf_mon_event *evt, int timeout)
{
	int mon_fd;
	static struct pollfd fds;
	struct sfi_request cmd;
	int ret;

	memset(&cmd, 0, sizeof(cmd));

	if (evt == NULL) {
		go_away(&cmd);
		return -1;
	}

	evt->type = EVT_INVALID;

	if ((mon_fd = prepare(&cmd, &fds)) == -1)
		return -1;

	if ((ret = poll(&fds, 1, timeout)) < 1) {
		return ret;
	}

	if (fds.revents & (POLLERR | POLLHUP)) {
		return 0;
	}

	if ((ret = recv(mon_fd, &cmd, sizeof(cmd), MSG_NOSIGNAL)) != sizeof(cmd)) {
		return 0;
	}

	switch (cmd.cmd) {
	case GETCPUINFO:
		evt->type = EVT_CPUINFO;
		memcpy(&evt->u.cpuinfo, &cmd.cpuinfo, sizeof(evt->u.cpuinfo));
		break;

	case GETPOLICY:
		evt->type = EVT_POLICY;
		memcpy(&evt->u.policy, &cmd.policy, sizeof(evt->u.policy));
		break;

	case SHUTDOWN:
		evt->type = EVT_SHUTDOWN;
		mon_fd = sf_set_mon_fd(-1);
		break;

	default:
		fprintf(stderr, "Unexpected monitor event: %d\n", cmd.cmd);
		break;
	}

	return 1;
}
