/*
 * CPU speed daemon - low-level Python client module
 *
 * Copyright 2003 Bryan O'Sullivan <bos@serpentine.com>
 *
 * This software is licensed under the GNU General Public License.
 * See COPYING for details.
 */

#include <Python.h>
#include "speedfreq.h"


static PyMethodDef pysf_methods[];
static PyObject *pysf_error;

static /* yes, this is disgusting */
#include "version.c"

void
init_speedfreq(void)
{
	PyObject *m, *d;

	m = Py_InitModule("_speedfreq", pysf_methods);
	d = PyModule_GetDict(m);
	pysf_error = PyErr_NewException("_speedfreq.error", NULL, NULL);
	PyDict_SetItemString(d, "error", pysf_error);

	PyDict_SetItemString(d, "version", Py_BuildValue("s", version));

	/* Policies. */

	PyDict_SetItemString(d, "POL_NONE", Py_BuildValue("i", SF_NILPOLICY));
	PyDict_SetItemString(d, "POL_POWERSAVE", Py_BuildValue("i", SF_POWERSAVE));
	PyDict_SetItemString(d, "POL_PERFORMANCE", Py_BuildValue("i", SF_PERFORMANCE));
	PyDict_SetItemString(d, "POL_DYNAMIC", Py_BuildValue("i", SF_DYNAMIC));
	PyDict_SetItemString(d, "POL_FIXED", Py_BuildValue("i", SF_FIXED));

	/* Events. */

	PyDict_SetItemString(d, "EVT_INVALID", Py_BuildValue("i", EVT_INVALID));
	PyDict_SetItemString(d, "EVT_SHUTDOWN", Py_BuildValue("i", EVT_SHUTDOWN));
	PyDict_SetItemString(d, "EVT_CPUINFO", Py_BuildValue("i", EVT_CPUINFO));
	PyDict_SetItemString(d, "EVT_POLICY", Py_BuildValue("i", EVT_POLICY));
}


static PyObject *
pysf_connect(PyObject *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, ":connect")) {
		return NULL;
	}

	int ret = sf_connect();

	if (ret == 1) {
		PyErr_SetString(pysf_error, "connect failed");
		return NULL;
	}

	Py_INCREF(Py_None);
	return Py_None;
}


static PyObject *
pysf_disconnect(PyObject *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, ":disconnect")) {
		return NULL;
	}

	sf_disconnect();

	Py_INCREF(Py_None);
	return Py_None;
}


static PyObject *
pysf_shutdown(PyObject *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, ":shutdown")) {
		return NULL;
	}

	if (sf_shutdown() != 0) {
		PyErr_SetString(pysf_error, "shutdown failed");
		return NULL;
	}

	Py_INCREF(Py_None);
	return Py_None;
}


static PyObject *
pysf_get_version(PyObject *self, PyObject *args)
{
	PyObject *ret = NULL;

	if (!PyArg_ParseTuple(args, ":get_version")) {
		goto done;
	}

	char *v = sf_getversion();

	if (v == NULL) {
		PyErr_SetString(pysf_error, "get_version failed");
		goto done;
	}

	ret = Py_BuildValue("s", v);
	free(v);

 done:
	return ret;
}


static PyObject *
pysf__get_policy(PyObject *self, PyObject *args)
{
	struct sf_policyinfo i;
	PyObject *ret;

	if (!PyArg_ParseTuple(args, ":_get_policy")) {
		return NULL;
	}

	if (sf_getpolicy(&i) != 0) {
		PyErr_SetString(pysf_error, "_get_policy failed");
		return NULL;
	}

	switch (i.policy) {
	case SF_DYNAMIC:
		ret = Py_BuildValue("iiii", i.policy,
				    i.dynamic.min, i.dynamic.max,
				    i.dynamic.current);
		break;
	case SF_FIXED:
		ret = Py_BuildValue("ii", i.policy, i.fixed.current);
		break;
	default:
		ret = Py_BuildValue("i", i.policy);
		break;
	}

	return ret;
}


static PyObject *
pysf__get_cpu_info(PyObject *self, PyObject *args)
{
	struct sf_cpuinfo i;

	if (!PyArg_ParseTuple(args, ":_get_cpu_info")) {
		return NULL;
	}

	if (sf_getcpuinfo(&i) != 0) {
		PyErr_SetString(pysf_error, "_get_cpu_info failed");
		return NULL;
	}

	return Py_BuildValue("iiisf", i.min, i.max, i.current, i.driver, i.idle);
}


static PyObject *
pysf_pol_powersave(PyObject *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, ":pol_powersave")) {
		return NULL;
	}

	if (sf_pol_powersave() != 0) {
		PyErr_SetString(pysf_error, "pol_powersave failed");
		return NULL;
	}

	Py_INCREF(Py_None);
	return Py_None;
}


static PyObject *
pysf_pol_performance(PyObject *self, PyObject *args)
{
	int r = sf_pol_performance();

	if (!PyArg_ParseTuple(args, ":pol_performance")) {
		return NULL;
	}

	if (r != 0) {
		PyErr_SetString(pysf_error, "pol_performance failed");
		return NULL;
	}

	Py_INCREF(Py_None);
	return Py_None;
}


static PyObject *
pysf_pol_dynamic(PyObject *self, PyObject *args)
{
	int min, max;

	if (!PyArg_ParseTuple(args, "ii:pol_dynamic", &min, &max)) {
		return NULL;
	}

	if (min < 0) {
		PyErr_SetString(PyExc_ValueError, "min must be positive");
		return NULL;
	}

	if (max < 0) {
		PyErr_SetString(PyExc_ValueError, "max must be positive");
		return NULL;
	}

	if (max <= min) {
		PyErr_SetString(PyExc_ValueError,
				"max must be greater than min");
		return NULL;
	}

	int r = sf_pol_dynamic(min, max);

	if (r != 0) {
		PyErr_SetString(pysf_error, "pol_dynamic failed");
		return NULL;
	}

	Py_INCREF(Py_None);
	return Py_None;
}


static PyObject *
pysf_pol_fixed(PyObject *self, PyObject *args)
{
	int freq;

	if (!PyArg_ParseTuple(args, "i:pol_fixed", &freq)) {
		return NULL;
	}

	if (freq <= 0) {
		PyErr_SetString(PyExc_ValueError, "freq must be positive");
		return NULL;
	}

	int r = sf_pol_fixed(freq);

	if (r != 0) {
		PyErr_SetString(pysf_error, "pol_dynamic failed");
		return NULL;
	}

	Py_INCREF(Py_None);
	return Py_None;
}


static PyObject *
pysf_get_mon_fd(PyObject *self, PyObject *args)
{
	if (!PyArg_ParseTuple(args, ":get_mon_fd")) {
		return NULL;
	}

	int fd = sf_get_mon_fd();

	if (fd == -1) {
		PyErr_SetString(pysf_error, "get_mon_fd failed: no monitoring in progress");
		return NULL;
	}

	return Py_BuildValue("i", fd);
}


static PyObject *
pysf__poll(PyObject *self, PyObject *args)
{
	struct sf_mon_event evt;
	PyObject *ret = NULL;
	int timeout;

	if (!PyArg_ParseTuple(args, "i:_poll", &timeout)) {
		goto done;
	}

	int r = sf_poll(&evt, timeout);

	if (r == 0) {
		ret = Py_None;
		Py_INCREF(ret);
		goto done;
	}

	if (r == -1) {
		PyErr_SetString(pysf_error, "poll failed");
		goto done;
	}

	switch (evt.type) {
	case EVT_CPUINFO:
	{
		const struct sf_cpuinfo *i = &evt.u.cpuinfo;

		ret = Py_BuildValue("iiiisf", evt.type, i->min, i->max,
				    i->current, i->driver, i->idle);
		break;
	}
	case EVT_POLICY:
	{
		const struct sf_policyinfo *i = &evt.u.policy;

		switch (i->policy) {
		case SF_DYNAMIC:
			ret = Py_BuildValue("iiiii", evt.type, i->policy,
					    i->dynamic.min, i->dynamic.max,
					    i->dynamic.current);
			break;
		case SF_FIXED:
			ret = Py_BuildValue("iii", evt.type, i->policy,
					    i->fixed.current);
			break;
		default:
			ret = Py_BuildValue("ii", evt.type, i->policy);
			break;
		}
		break;
	}
	case EVT_SHUTDOWN:
		ret = Py_BuildValue("i", evt.type);
		break;
	case EVT_INVALID:
		break;
	}

 done:
	return ret;
}


static PyMethodDef pysf_methods[] = {
	{ "connect", pysf_connect, METH_VARARGS,
	  "Connect to the speedfreq daemon." },
	{ "disconnect", pysf_disconnect, METH_VARARGS,
	  "Disconnect from the speedfreq daemon." },
	{ "shutdown", pysf_shutdown, METH_VARARGS,
	  "Ask the speedfreq daemon to shut down." },
	{ "get_version", pysf_get_version, METH_VARARGS,
	  "Return a string containing the server's version information." },
	{ "_get_policy", pysf__get_policy, METH_VARARGS,
	  "Return a tuple describing the current CPU speed policy." },
	{ "_get_cpu_info", pysf__get_cpu_info, METH_VARARGS,
	  "Return a tuple describing the CPU." },
	{ "pol_powersave", pysf_pol_powersave, METH_VARARGS,
	  "Set the CPU speed policy to save power." },
	{ "pol_performance", pysf_pol_performance, METH_VARARGS,
	  "Set the CPU speed policy to maximise performance." },
	{ "pol_dynamic", pysf_pol_dynamic, METH_VARARGS,
	  "Set the CPU speed policy to adjust dynamically." },
	{ "pol_fixed", pysf_pol_fixed, METH_VARARGS,
	  "Set the CPU speed policy to a fixed speed." },
	{ "get_mon_fd", pysf_get_mon_fd, METH_VARARGS,
	  "Return the fd used to get monitoring data from the daemon." },
	{ "_poll", pysf__poll, METH_VARARGS,
	  "Poll for an event notification from the daemon." },
	{ NULL, NULL, 0, NULL },
};
