/*
 * (Quick hack in the shape of an) Automatic CPU speed governer.
 *
 * Measures the system's current idleness by reading /proc/uptime.  If
 * the system goes above 80% idle, the target CPU speed is dropped by
 * 100MHz.  If it goes below 30% idle, CPU speed is increased by
 * 100MHz.
 *
 * The daemon starts polling /proc/uptime every 300ms.  If the target
 * frequency is unchanged for 10 samples, the sample rate is dropped
 * to 1s.  It reverts to 300ms when there's a change.
 *
 * This now implements a client-server system.  If a server doesn't
 * already exist, speedfreq will become a server, listening to requests
 * on a socket.  Subsequent instances of speedfreq talk to the existing
 * server to do their work.  This allows users to change and query the
 * current policy.
 *
 * TODO:
 *   - have an adjustable ceiling for preferring power savings over performance
 *   - predicates for when to choose power savings:
 *     - running off battery
 *     - estimated running time remaining
 *     - thermal conditions
 *   - config file? command-line opts are easier
 *   - SMP support?
 *
 * Copyright 2003 Jeremy Fitzhardinge <jeremy@goop.org>
 *
 * This software is licensed under the GNU General Public License.
 * See COPYING for details.
 *
 * $Id: //depot/cpuspeed/speedfreq.c#4 $
 */

#define _XOPEN_SOURCE 500	/* for pread() */
#define _BSD_SOURCE		/* for daemon() */

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>

#include <sys/socket.h>
#include <sys/un.h>

#include "speedfreq.h"

extern const char version[];

#define NAME		"speedfreq"

/*
 * Tunables
 */

static const char *progname;

static int clo_debug;		/* debug messages */

static void mon_cpuinfo(const struct sf_cpuinfo *c)
{
	printf("CPU new speed %d MHz, %g%% idle\n",
	       c->current / 1000, c->idle * 100.);
}

#define POLICYLEN	64

static void policy_str(const struct sf_policyinfo *p, char *buf)
{
	switch(p->policy) {
	case SF_POWERSAVE:
	case SF_PERFORMANCE:
		sprintf(buf, "%s", sf_policyname(p->policy));
		break;
	case SF_FIXED:
		sprintf(buf, "fixed %d MHz\n", 
			p->fixed.current / 1000);
		break;
	case SF_DYNAMIC:
		sprintf(buf, "dynamic %d-%d MHz", 
			p->dynamic.min / 1000, p->dynamic.max / 1000);
		break;
	case SF_NILPOLICY:
		sprintf(buf, "none?");
		break;
	}	
}

static void mon_policy(const struct sf_policyinfo *p)
{
	char buf[POLICYLEN];

	policy_str(p, buf);
	printf("New policy: %s\n", buf);
}

int main(int argc, char **argv)
{
	int opt;
	int usage = 0;
	int doquit = 0;
	int getstatus = 0;
	int getcpuinfo = 0;
	int showversion = 0;
	int domonitor = 0;

	const char *newpolicy = NULL;
	enum sf_policy policy = SF_NILPOLICY;
	int pol_arg = -1;
	int ret;

	progname = argv[0];

	while((opt = getopt(argc, argv, "DVcmp:qs?")) != EOF) {
		switch(opt) {
		case 'c':
			getcpuinfo = 1;
			break;

		case 'm':
			domonitor = 1;
			break;

		case 'p':
			newpolicy = optarg;
			break;

		case 'q':
			doquit = 1;
			break;

		case 's':
			getstatus = 1;
			break;

		case 'D':
			clo_debug = 1;
			break;

		case 'V':
			showversion = 1;
			break;

		case '?':
			usage = 1;
			break;

		default:
			usage = 2;
			break;
		}
	}

	if (optind != argc)
		usage = 2;


	if (newpolicy != NULL) {
		long l;
		char *end;

		if (strcmp(newpolicy, "powersave") == 0)
			policy = SF_POWERSAVE;
		else if (strcmp(newpolicy, "performance") == 0)
			policy = SF_PERFORMANCE;
		else if (strcmp(newpolicy, "dynamic") == 0)
			policy = SF_DYNAMIC;
		else if ((l = strtol(newpolicy, &end, 10)) && *end == '\0') {
			policy = SF_FIXED;
			pol_arg = (int)l * 1000; /* MHz -> kHz */
		} else {
			fprintf(stderr, "%s: unknown policy \"%s\"\n",
				progname, newpolicy);
			usage = 2;
		}
	}

	if (usage) {
		printf("Usage: %s [-cmqsDV] [-p policy]\n", progname);
		printf("\t-c\tget info about CPU\n"
		       "\t-m\tmonitor daemon for policy and CPU speed changes\n"
		       "\t-p P\tset policy to P\n"
		       "\t-q\ttell server to quit\n"
		       "\t-s\tget current status\n"
		       "\t-D\tprint debug messages\n"
		       "\t-V\tversion information\n"

		       "\nAvailable policies:\n"
		       "\tpowersave\tmaximise power savings\n"
		       "\tperformance\tmaximise performance\n"
		       "\tdynamic\t\tadjust speed according to CPU load (default)\n"
		       "\tNNN\t\tset CPU to a fixed speed of NNN MHz\n"
		       "\n"
			);
		exit(usage == 2 ? 1 : 0);
	}


	if ((ret = sf_connect()))
		goto out;

	/* show status if given nothing else to do */
	if (!getstatus &&
	    policy == SF_NILPOLICY &&
	    !doquit &&
	    !showversion &&
	    !getcpuinfo &&
	    !domonitor)
		getstatus = 1;

	if (policy != SF_NILPOLICY) {
		/* client */
		switch(policy) {
		case SF_POWERSAVE:	ret = sf_pol_powersave(); break;
		case SF_PERFORMANCE:	ret = sf_pol_performance(); break;
		case SF_DYNAMIC:	ret = sf_pol_dynamic(0, 0); break;
		case SF_FIXED:		ret = sf_pol_fixed(pol_arg); break;
		case SF_NILPOLICY:	abort();
		}
	}

	if (getstatus) {
		struct sf_policyinfo info;		
		ret = sf_getpolicy(&info);

		if (ret == 0) {
			char buf[POLICYLEN];
			
			policy_str(&info, buf);
			printf("Current policy is %s\n", buf);
		}
	}

	if (getcpuinfo) {
		struct sf_cpuinfo c;

		ret = sf_getcpuinfo(&c);

		if (ret == 0) {
			printf("CPU speed: min %dMHz, "
			       "max %dMHZ, "
			       "current %dMHz; %.2f%% idle\n"
			       "CPUfreq driver: %s\n",
			       c.min/1000, c.max/1000, c.current/1000,
			       c.idle * 100,
			       c.driver);
		}
	}

	if (showversion) {
		char *v;
		printf("Client version: %s\n", version);

		v = sf_getversion();
		if (v != NULL) {
			printf("Server version: %s\n", v);
			free(v);
		}
	}

	if (domonitor) {
		static const struct sf_monitorcb cb = {
			.cpuinfo = mon_cpuinfo,
			.policy = mon_policy,
		};

		while(sf_monitor(&cb, 1) != -1)
			;
	}

	if (doquit) {
		ret = sf_shutdown();
	}

  out:
	return ret;
}
