# CPU speed daemon - high-level Python client module
#
# Copyright 2003 Bryan O'Sullivan <bos@serpentine.com>
#
# This software is licensed under the GNU General Public License.
# See COPYING for details.

"""Python interface to the speedfreq CPU speed management daemon.

Policy-related constants:

    POL_NONE        no policy is set
    POL_POWERSAVE   save power at the expense of performance
    POL_PERFORMANCE gain performance at the expense of power
    POL_FIXED       run the CPU at a fixed performance level
    POL_DYNAMIC     vary CPU performance in response to demand

Event-related constants:

    EVT_INVALID     this event contains no valid information
    EVT_SHUTDOWN    the speedfreq daemon has shut down
    EVT_CPUINFO     CPU info has changed
    EVT_POLICY      the speed management policy has changed

Other constants:

    version         version of the client API

Policy-related classes:

    Policy          base policy class
    PowerSave
    Performance
    Fixed
    Dynamic

Event-related classes:

    Event           base event class
    CpuInfoEvent
    ShutdownEvent
    PowerSaveEvent
    PerformanceEvent
    FixedEvent
    DynamicEvent

Functions:

    connect         connect to the speedfreq daemon
    disconnect      disconnect from the speedfreq daemon
    get_version     get the version of the speedfreq daemon
    shutdown        ask the speedfreq daemon to shut down
    get_policy      get the current speed management policy
    get_cpu_info    get information about the CPU
    set_policy      set the speed management policy
    poll            poll for speed management events
    get_mon_fd      return the fd used by poll()
"""

__author__ = "Bryan O'Sullivan <bos@serpentine.com>"
__version__ = "$Revision: 1.14 $"


from _speedfreq import *
import _speedfreq as _sf


class Policy(object):
    '''Base class defining policy information.

    Instance attributes:

        policy      numeric policy descriptor (one of the POL_* constants)
    '''

    def __init__(self, num):
        self.policy = num

    def __repr__(self):
        return '<Policy %d>' % self.policy


class PowerSave(Policy):
    '''Power-saving policy.
    Attempt to extend battery life at the expense of performance.'''

    def __init__(self):
        super(PowerSave, self).__init__(POL_POWERSAVE)

    def __repr__(self):
        return '<PowerSave>'


class Performance(Policy):
    '''Performance policy.
    Attempt to improve performance, at the expense of battery life.'''

    def __init__(self):
        super(Performance, self).__init__(POL_PERFORMANCE)

    def __repr__(self):
        return '<Performance>'


class Dynamic(Policy):
    '''Dynamic speed management policy.
    Vary CPU speed according to load.

    Instance attributes:

        min         minimum frequency
        max         maximum frequency
        current     current frequency (None if not applicable)
    '''

    def __init__(self, min, max, current = None):
        super(Dynamic, self).__init__(POL_DYNAMIC)
        self.min, self.max, self.current = min, max, current

    def __repr__(self):
        return ('<Dynamic range=[%d, %d] cur=%d>' %
                (self.min, self.max, self.current))


class Fixed(Policy):
    '''Fixed policy.
    Keep the CPU running at a fixed rate.

    Instance attributes:

        freq        operating frequency
    '''

    def __init__(self, freq):
        super(Fixed, self).__init__(POL_FIXED)
        self.freq = freq

    def __repr__(self):
        return ('<Fixed cur=%d>' % self.freq)


_policy_map = {
    POL_POWERSAVE: PowerSave,
    POL_PERFORMANCE: Performance,
    POL_DYNAMIC: Dynamic,
    POL_FIXED: Fixed,
    }


def _dispatch(map, args):
    try:
        args = list(args)
    except TypeError:
        args = [args]
    return map[args.pop(0)](*args)


def get_policy():
    '''Return a Policy object describing the current CPU speed policy.'''

    return _dispatch(_policy_map, _sf._get_policy())


class CpuInfo(object):
    '''Describe the CPU\'s speed management capabilities.

    Instance attributes:

        min         the minimum frequency supported
        max         the maximum frequency supported
        current     the current operating frequency
        driver      the cpufreq driver used by the kernel
        idle        the idleness of the CPU, between 0 and 1.0
    '''

    def __init__(self, args):
        self.min, self.max, self.current, self.driver, self.idle = args

    def __repr__(self):
        return ('<CpuInfo %s range=[%d, %d] cur=%d idle=%.2g>' %
                (self.driver, self.min, self.max, self.current, self.idle))


def get_cpu_info():
    '''Return a CpuInfo object describing the CPU.'''

    return CpuInfo(_sf._get_cpu_info())


def set_policy(pol, **kwargs):
    '''Change the CPU speed management policy to POL.
    POL can be one of the POL_* constants, a Policy class, or a Policy
    object (really, an object that has the appropriate attributes).

    If POL is a constant or a class, rather than an object, KWARGS may
    need to contain some appropriate keys.

    If the policy is fixed, KWARGS must have a "freq" key, or else the
    default will be the minimum frequency supported by the CPU.

    If the policy is dynamic, KWARGS must have "min" and "max" keys,
    or else the defaults will be the minimum and maximum frequencies
    supported by the CPU.

    A few examples:

        import speedfreq as sf

        sf.set_policy(POL_POWERSAVE)
        sf.set_policy(sf.Performance)
        sf.set_policy(POL_FIXED, freq = 666)
        sf.set_policy(sf.Dynamic(min = 800, max = 2400))

    My, how accommodatingly flexible we are.
    '''

    i = get_cpu_info()
    if pol == POL_FIXED or hasattr(pol, 'freq') or pol is Fixed:
        pol_fixed(getattr(pol, 'freq', kwargs.get('freq', i.min)))
    elif (pol == POL_DYNAMIC or hasattr(pol, 'min') or hasattr(pol, 'max') or
          pol is Dynamic):
        pol_dynamic(getattr(pol, 'min', kwargs.get('min', i.min)),
                    getattr(pol, 'max', kwargs.get('max', i.max)))
    elif (pol == POL_PERFORMANCE or isinstance(pol, Performance) or
          pol is Performance):
        pol_performance()
    elif (pol == POL_POWERSAVE or isinstance(pol, PowerSave) or
          pol is PowerSave):
        pol_powersave()
    else:
        raise TypeError('invalid policy')


class Event(object):
    '''The base event class.
    Subclasses detail CPU speed management events of various kinds.

    Instance attributes:

        type        numeric event descriptor (one of the EVT_* constants)
    '''

    def __init__(self, type):
        self.type = type

    def __repr__(self):
        return '<Event %d>' % self.type


class ShutdownEvent(Event):
    '''The speedfreq daemon is shutting down.'''

    def __init__(self):
        super(ShutdownEvent, self).__init__(EVT_SHUTDOWN)

    def __repr__(self):
        return '<ShutdownEvent>'


# This stuff causes a bunch of event classes to come into existence.

class _PolicyEventClass(type):
    def __new__(cls, name, bases, dict):
        return type.__new__(cls, ''.join([b.__name__ for b in bases]),
                            bases, dict)

def _policy_event(policy):
    class PolicyEvent(policy, Event):
        __metaclass__ = _PolicyEventClass
        def __init__(self, *args):
            super(Event, self).__init__(EVT_POLICY)
            policy.__init__(self, *args)
        def __repr__(self):
            return policy.__repr__(self)
    cls = PolicyEvent
    globals()[cls.__name__] = cls
    return cls

_pol_event_map = {
    POL_POWERSAVE: _policy_event(PowerSave),
    POL_PERFORMANCE: _policy_event(Performance),
    POL_DYNAMIC: _policy_event(Dynamic),
    POL_FIXED: _policy_event(Fixed),
    }


class CpuInfoEvent(CpuInfo, Event):
    '''Some aspect of the CPU (for example, its speed) has changed.'''

    def __init__(self, args):
        Event.__init__(self, EVT_CPUINFO)
        CpuInfo.__init__(self, args)

    def __repr__(self):
        return CpuInfo.__repr__(self)


_event_map = {
    EVT_CPUINFO: CpuInfoEvent,
    EVT_POLICY: lambda ret: _dispatch(_pol_event_map, ret),
    EVT_SHUTDOWN: ShutdownEvent,
    }


def poll(timeout = -1):
    '''Block for up to TIMEOUT milliseconds until an event comes in.
    If an event comes in, return an Event object, otherwise None.

    If TIMEOUT is -1 (the default), block forever.  If 0, do not
    block, but return immediately whether or not there\'s an event
    ready.'''

    ret = _sf._poll(timeout)
    if ret is None: return ret
    ret = list(ret)
    return _event_map[ret.pop(0)](ret)


def _test():
    import time
    connect()

    i = get_cpu_info()
    print 'CPU info:', i
    p = get_policy()
    print 'Initial policy:', p

    def test_policy(desc, func):
        print 'Testing %s policy' % desc
        func()
        printed = False
        while True:
            evt = poll(10)
            if evt is None: break
            if not printed:
                print 'Results:'
                printed = True
            print '   ', evt
        time.sleep(0.5)

    try:
        poll(0)
        print 'Monitoring file descriptor:', get_mon_fd()

        test_policy('performance I', pol_performance)

        test_policy('powersave I', pol_powersave)

        delta = (i.max - i.min) / 3

        test_policy('fixed I', lambda: pol_fixed(i.min + delta))

        test_policy('dynamic I',
                    lambda: pol_dynamic(i.min + delta, i.max - delta))

        test_policy('performance II',
                    lambda: set_policy(Performance()))

        test_policy('powersave II',
                    lambda: set_policy(PowerSave()))

        test_policy('fixed II',
                    lambda: set_policy(Fixed(i.min + delta)))

        test_policy('dynamic II',
                    lambda: set_policy(Dynamic(i.min + delta, i.max - delta)))

        test_policy('performance III',
                    lambda: set_policy(Performance))

        test_policy('powersave III',
                    lambda: set_policy(PowerSave))

        test_policy('fixed III',
                    lambda: set_policy(Fixed, freq = i.min + delta))

        test_policy('dynamic III',
                    lambda: set_policy(Dynamic, min = i.min + delta,
                                       max = i.max - delta))

        start = time.time()
        run_time = 5
        end = start + run_time
        now = start

        import os
        pid = os.fork()
        if pid == 0:
            time.sleep(1)
            end = start + run_time / 2.0
            while now < end:
                now = time.time()
            os._exit(0)

        print 'Waiting %d secs for events' % run_time

        while end - now > 0.1:
            evt = poll((end - now) * 1000)
            print 'Event:', evt
            now = time.time()

    finally:
        time.sleep(1)
        print 'Restoring policy to its initial state'
        set_policy(p)

    disconnect()


if __name__ == '__main__':
    _test()

del _test
